from netgen.csg import *
from ngsolve import *

ngsglobals.msg_level = 2
SetHeapSize(10000000)

# -----------------------  PARAMETERS -----------------------
h_max = 0.11					# Global mesh size
k = 2							# Order of velocity space
gd = 0.1						# Grad-div stbilisation parameter

dt = 1/100						# Time step
t = Parameter(0.0)				# Time variable
t_end = dt 						# End time

nu = 1e-3						# Viscosity
U = 0.45						# Inflow speed

pReg = 1e-8						# Pressure regularisation parameter
inverse = "sparsecholesky"		# Solver for linear systems
max_it = 10						# Maximum number of Richardson itterations
tol = 1e-8						# Residula tolerance for Richardson
print_resid=False				# Print Richardson residuals

# ------------------------- MESH -------------------------
channel = OrthoBrick( Pnt(0.0,0.0,0.0), Pnt(2.5,0.41,0.41) )
geo = CSGeometry()
geo.Add ( channel )

ngmesh = geo.GenerateMesh(maxh=h_max)

ngmesh.SetBCName(0, "wall")
ngmesh.SetBCName(1, "inflow")
ngmesh.SetBCName(2, "wall")
ngmesh.SetBCName(3, "outflow")
ngmesh.SetBCName(4, "wall")
ngmesh.SetBCName(5, "wall")

mesh = Mesh(ngmesh)

## Check Mesh
# from ngsolve.internal import *
# boundaries = ["wall","inflow","outflow"]

# fes = H1(mesh)
# u = GridFunction(fes,"u")
# viewoptions.clipping.enable = 1
# Draw(u)
# visoptions.clipsolution = "scal"
# for bc in boundaries:
#     u.Set(CoefficientFunction(1),definedon=mesh.Boundaries(bc))
#     Redraw()
#     input(bc)

# -----------------------  BOUNDARY DATA -----------------------
uInflow = CoefficientFunction( (16*U*(y*(0.41-y)/(0.41**2))*(z*(0.41-z)/(0.41**2)), 0.0, 0.0) )

# ----------------------- FE SPACE -----------------------
V = VectorH1(mesh, order=k, dirichlet="wall|inflow")
Q = H1(mesh, order=k-1)
X = FESpace([V,Q])

gfu = GridFunction(X)
vel = gfu.components[0]				# Velocity Component
pre = gfu.components[1]				# Pressure Component

Draw(vel, mesh, "vel")
Draw(pre, mesh, "pre")

# ----------------------- FE METHOD -----------------------
(u,p), (v,q) = X.TnT()

Stokes = nu*InnerProduct(grad(u),grad(v)) - p*div(v) - q*div(u) + gd*div(u)*div(v)
Convect = InnerProduct(grad(u)*u,v)

## Matrices
mStar = BilinearForm(X, symmetric=True)
mStar += SymbolicBFI(form=InnerProduct(u,v))
mStar += SymbolicBFI(form=dt*Stokes )
with TaskManager():
	mStar.Assemble()

mStarReg = BilinearForm(X, symmetric=True)
mStarReg += SymbolicBFI(form=InnerProduct(u,v))
mStarReg += SymbolicBFI(form=dt*Stokes - dt*pReg*p*q)
with TaskManager():
	mStarReg.Assemble()
	inv_mStarReg = mStarReg.mat.Inverse(freedofs=X.FreeDofs(), inverse=inverse)

Conv = BilinearForm(X, symmetric=False, nonassemble=True)
Conv += SymbolicBFI(form=Convect).SetIntegrationRule(TRIG,IntegrationRule(TRIG,3*k-1))

m = BilinearForm(X, symmetric=True)
m += SymbolicBFI(form=InnerProduct(u,v))
with TaskManager():
	m.Assemble()

# ----------------------- SOLVE PROBLEM -----------------------
res = gfu.vec.CreateVector()

with TaskManager():
	for it in range(1,int(t_end/dt + 1)):
		t.Set(it*dt)
		print("t={:1.8f}".format(t.Get()), end="\r")

		Conv.Apply(gfu.vec, res)
		res.data = -dt*res
		res.data += m.mat*gfu.vec

		gfu.components[0].Set(uInflow, definedon=mesh.Boundaries("inflow"))
		Redraw()
		input("inflow")
		res.data += -mStar.mat*gfu.vec 

		gfu.vec.data += solvers.PreconditionedRichardson(a=mStar, rhs=res, pre=inv_mStarReg, freedofs=X.FreeDofs(), 
															maxit=max_it, tol=tol, dampfactor=1.0, printing=print_resid)

		Redraw(blocking=True)